/*
 * Neural network stuff
 * Written by Bartek Kania, while terribly bored
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include "nnet.h"

static void nn_init_units(struct nn_layer *nl)
{
    struct nn_unit *unit;
    int i, j;

    srand(time(0));

    for (i = 0; i < nl->our_units; i++) {
	unit = &nl->units[i];
	unit->value = 0;
	unit->thresh = (0.2 * (rand()/(RAND_MAX+1.0))) - 0.1;
	unit->weights = calloc(nl->next_units, sizeof(double));
	for (j = 0; j < nl->next_units; j++)
	    unit->weights[j] = (0.2 * (rand()/(RAND_MAX+1.0))) - 0.1;
    }
}

struct nn_net *nn_init(int layers, int *sizes)
{
    struct nn_net *nnet;
    struct nn_layer *nl;
    int i;

    if (layers < 2)
	return 0;

    nnet = malloc(sizeof(struct nn_net));
    nnet->num_layers = layers;
    nnet->input_units = sizes[0];
    nnet->output_units = sizes[layers - 1];

    nl = nnet->layers = calloc(layers, sizeof(struct nn_net));

    for (i = 0; i < layers; i++) {
	nl[i].next_units = 0;
	if (i) 
	    nl[i].prev_units = sizes[i - 1];
	if ((i + 1) < layers)
	    nl[i].next_units = sizes[i + 1];
	nl[i].our_units = sizes[i];
	nl[i].units = calloc(sizes[i], sizeof(struct nn_unit));
	nn_init_units(&nl[i]);
    }

    return nnet;
}

static void nn_free_layer(struct nn_layer *nl)
{
    int i;

    for (i = 0; i < nl->our_units; i++)
	free(nl->units[i].weights);
    free(nl->units);
}

void nn_free(struct nn_net *nnet)
{
    int i;

    for (i = 0; i < nnet->num_layers; i++)
	nn_free_layer(&nnet->layers[i]);

    free(nnet->layers);
}

static inline double nn_sigmoid(double in)
{
    return 1.0/(1.0 + exp(-in));
}

static void nn_feedforward(struct nn_layer *src, struct nn_layer *dst)
{
    int i, j;
    double t;

    /* Firstly, clear out all junk data */
    for (i = 0; i < dst->our_units; i++)
	dst->units[i].value = dst->units[i].thresh;

    /* For each unit, send the stuff out to all followers */
    for (i = 0; i < src->our_units; i++) {
	t = src->units[i].value;
	for (j = 0; j < dst->our_units; j++)
	    dst->units[j].value += t * src->units[i].weights[j];
    }

    /* Now, calculate the activation-function for each unit */
    for (i = 0; i < dst->our_units; i++)
	dst->units[i].value = nn_sigmoid(dst->units[i].value);
}

double *nn_compute(struct nn_net *nnet, double *input)
{
    int i;
    struct nn_layer *nl;
    double *output;

    /* Start by setting the input in the input-layer */
    nl = &nnet->layers[0];
    for (i = 0; i < nl->our_units; i++)
	nl->units[i].value = input[i];

    /* Run the data through the net */
    for (i = 0; i < (nnet->num_layers - 1); i++)
	nn_feedforward(&nnet->layers[i], &nnet->layers[i+1]);

    /* Output the results */
    nl = &nnet->layers[nnet->num_layers - 1];
    if (!(output = calloc(nl->our_units, sizeof(double)))) 
	return 0;

    for (i = 0; i < nl->our_units; i++)
	output[i] = nl->units[i].value;

    return output;
}


/* This moves backwards through the layers. The weights leading to the 'curr'
 * layer are updated, and then the error-rates for the previous layer are
 * computed */
static void nn_backpropagate(struct nn_layer *curr, struct nn_layer *prev, 
			     double rate)
{
    int i, j;
    double wchange, err;

    /* First, fix the threshhold value. */
    for (i = 0; i < curr->our_units; i++) 
	curr->units[i].thresh += rate * curr->units[i].error;

    /* Calculate the error in the previos layer */
    for (i = 0; i < prev->our_units; i++) {
	err = 0;
	for (j = 0; j < curr->our_units; j++) 
	    err += curr->units[j].error * prev->units[i].weights[j];
	err *= prev->units[i].value * (1 - prev->units[i].value);
	prev->units[i].error = err;
    }

    /* Update the weights for the previous layer */
    for (i = 0; i < prev->our_units; i++) {
        for (j = 0; j < curr->our_units; j++) {
	    wchange = rate * curr->units[j].error * prev->units[i].value;
	    prev->units[i].weights[j] += wchange;
	}
    }
		
}

double nn_teach(struct nn_net *nnet, double *input, double *output, 
		double rate)
{
    double err = 0, tmp, *junk;
    int i;
    struct nn_layer *curr, *prev;

    /* Feed the data through the net */
    junk = nn_compute (nnet, input);
    free (junk);

    curr = &nnet->layers[nnet->num_layers - 1];

    /* Clean up the output nodes */
    for (i = 0; i < curr->our_units; i++) {
	if (curr->units[i].value > TRUE_LIMIT)
	    curr->units[i].value = 1;
	else if (curr->units[i].value < FALSE_LIMIT)
	    curr->units[i].value = 0;
    }

    /* Start by calculating the error in the output nodes */
    for (i = 0; i < curr->our_units; i++) {
	curr->units[i].error = curr->units[i].value * 
	    (1 - curr->units[i].value) * (output[i] - curr->units[i].value);
	tmp = output[i] - curr->units[i].value;
	err += tmp * tmp;
    }

    /* Backpropagate, and update the weights */
    for (i = nnet->num_layers - 1; i; i--) {
	curr = &nnet->layers[i];
	prev = &nnet->layers[i - 1];
	nn_backpropagate (curr, prev, rate);
    }

    return err;
}
